# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

Workflows for pyspread

"""

from ast import literal_eval
from base64 import b85encode
import bz2
from contextlib import contextmanager
from copy import copy
import csv
import io
import logging
import numpy
import os.path
from pathlib import Path
from shutil import move
from tempfile import NamedTemporaryFile
from typing import Iterable, Tuple

from PyQt6.QtCore import (Qt, QMimeData, QModelIndex, QBuffer, QRect, QRectF,
                          QItemSelectionModel, QSize)
from PyQt6.QtGui import QTextDocument, QImage, QPainter, QUndoCommand
from PyQt6.QtWidgets import (QApplication, QMessageBox, QInputDialog,
                             QStyleOptionViewItem, QTableView)
try:
    from PyQt6.QtSvg import QSvgGenerator
except ImportError:
    QSvgGenerator = None

try:
    import matplotlib
    import matplotlib.figure as matplotlib_figure
except ImportError:
    matplotlib = None
    matplotlib_figure = None

try:
    import openpyxl
except ImportError:
    openpyxl = None

try:
    from pyspread import commands
    from pyspread.dialogs \
        import (DiscardChangesDialog, FileOpenDialog, GridShapeDialog,
                FileSaveDialog, ImageFileOpenDialog, ChartDialog,
                CellKeyDialog, FindDialog, ReplaceDialog, CsvFileImportDialog,
                CsvImportDialog, CsvExportDialog, CsvExportAreaDialog,
                FileExportDialog, SvgExportAreaDialog, SinglePageArea)
    from pyspread.interfaces.pys import PysReader, PysWriter
    from pyspread.interfaces.xlsx import XlsxReader
    from pyspread.lib.attrdict import AttrDict
    from pyspread.lib.hashing import sign, verify
    from pyspread.lib.selection import Selection
    from pyspread.lib.typechecks import is_svg, check_shape_validity
    from pyspread.lib.csv import csv_reader, convert
    from pyspread.lib.file_helpers import \
        (linecount, file_progress_gen, ProgressDialogCanceled)
    from pyspread.model.model import CellAttribute, class_format_functions
except ImportError:
    import commands
    from dialogs \
        import (DiscardChangesDialog, FileOpenDialog, GridShapeDialog,
                FileSaveDialog, ImageFileOpenDialog, ChartDialog,
                CellKeyDialog, FindDialog, ReplaceDialog, CsvFileImportDialog,
                CsvImportDialog, CsvExportDialog, CsvExportAreaDialog,
                FileExportDialog, SvgExportAreaDialog, SinglePageArea)
    from interfaces.pys import PysReader, PysWriter
    from interfaces.xlsx import XlsxReader
    from lib.attrdict import AttrDict
    from lib.hashing import sign, verify
    from lib.selection import Selection
    from lib.typechecks import is_svg, check_shape_validity
    from lib.csv import csv_reader, convert
    from lib.file_helpers import \
        (linecount, file_progress_gen, ProgressDialogCanceled)
    from model.model import CellAttribute, class_format_functions


class Workflows:
    """Workflow container class"""

    cell2dialog = {}  # Stores acrive chart dialogs

    def __init__(self, main_window):
        self.main_window = main_window

    @contextmanager
    def busy_cursor(self):
        """:class:`~contextlib.contextmanager` that displays a busy cursor"""

        QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
        yield
        QApplication.restoreOverrideCursor()

    @contextmanager
    def prevent_updates(self):
        """:class:`~contextlib.contextmanager` sets the prevent_updates state

        The prevent_updates state prevents updates in main_window.grid.setData

        """

        self.main_window.prevent_updates = True
        yield
        self.main_window.prevent_updates = False

    def handle_changed_since_save(func, *args, **kwargs):
        """Decorator to handle changes since last saving the document

        If changes are present then a dialog is displayed that asks if the
        changes shall be discarded.

        - If the user selects `Cancel` then `func` is not executed.
        - If the user selects `Save` then the file is saved and `func` is
          executed.
        - If the user selects `Discard` then the file is not saved and `func`
          is executed.

        If no changes are present then `func` is directly executed.
        After executing `func`, :func:`reset_changed_since_save` and
        `update_main_window_title` are called.

        """

        def function_wrapper(self, *args, **kwargs):
            """Check changes and display and handle the dialog"""

            if self.main_window.settings.changed_since_save:
                choice = DiscardChangesDialog(self.main_window).choice
                if choice is None:
                    return
                if not choice:
                    # We try to save to a file
                    if self.file_save() is False:
                        # File could not be saved --> Abort
                        return
            try:
                func(self, *args, **kwargs)
            except TypeError:
                func(self)  # No args accepted
            self.reset_changed_since_save()
            self.update_main_window_title()

        return function_wrapper

    def reset_changed_since_save(self):
        """Sets changed_since_save to False and updates the window title"""

        # Change the main window filepath state
        self.main_window.settings.changed_since_save = False

    def update_main_window_title(self):
        """Change the main window title to reflect the current file name"""

        # Get the current filepath
        filepath = self.main_window.settings.last_file_input_path
        if filepath == Path.home():
            title = "pyspread"
        else:
            title = f"{filepath.name} - pyspread"
        self.main_window.setWindowTitle(title)

    @handle_changed_since_save
    def file_new(self):
        """File new workflow"""

        # Get grid shape from user
        old_shape = self.main_window.grid.model.code_array.shape
        shape = GridShapeDialog(self.main_window, old_shape).shape

        # Check if shape is valid

        if shape is None:
            return

        try:
            check_shape_validity(shape, self.main_window.settings.maxshape)
        except ValueError as err:
            self.main_window.statusBar().showMessage('Error: ' + str(err))
            return

        # Set current cell to upper left corner
        for grid in self.main_window.grids:
            grid.current = 0, 0, 0
            # Select upper left cell because initial selection behaves strange
            grid.reset_selection()

        # Reset grid
        self.main_window.grid.model.reset()

        # Empty undo stack
        self.main_window.undo_stack.clear()

        # Delete old filepath
        self.main_window.settings.last_file_input_path = Path.home()

        # Set new shape
        self.main_window.grid.model.shape = shape

        # Update cell spans and zoom because this is unsupported by the model
        for grid in self.main_window.grids:
            with grid.undo_resizing_row():
                with grid.undo_resizing_column():
                    grid.update_cell_spans()
                    grid.update_zoom()

            # Update index widgets
            grid.update_index_widgets()

        # Set current cell to upper left corner
        for grid in self.main_window.grids:
            grid.current = 0, 0, 0

        # Change the main window filepath state
        self.main_window.settings.changed_since_save = False

        # Update macro editor
        self.main_window.macro_panel.update_()

        # Exit safe mode
        self.main_window.safe_mode = False

    def count_file_lines(self, filepath: Path):
        """Returns line count of file in filepath

        :param filepath: Path of file to be analyzed

        """

        try:
            with open(filepath, 'rb') as infile:
                return linecount(infile)
        except OSError as error:
            self.main_window.statusBar().showMessage(str(error))
            return

    def filepath_open(self, filepath: Path):
        """Workflow for opening a file if a filepath is known

        :param filepath: Path of file to be opened

        """

        grid = self.main_window.grid
        code_array = grid.model.code_array

        # Get number of lines for progress dialog
        filelines = self.count_file_lines(filepath)
        if not filelines:  # May not be None or 0
            return

        # Reset grid
        grid.model.reset()

        # Empty undo stack
        self.main_window.undo_stack.clear()

        # Reset macro editor
        self.main_window.macro_panel.macro_editor.clear()

        # Is the file signed properly ?
        self.main_window.safe_mode = True
        signature_key = self.main_window.settings.signature_key
        try:
            with open(filepath, "rb") as infile:
                signature_path = filepath.with_suffix(filepath.suffix + '.sig')
                with open(signature_path, "rb") as sigfile:
                    self.main_window.safe_mode = not verify(infile.read(),
                                                            sigfile.read(),
                                                            signature_key)
        except OSError:
            self.main_window.safe_mode = True

        # File format handling
        if filepath.suffix == ".pysu":
            fopen = open
            freader = PysReader
        elif filepath.suffix == ".pys":
            fopen = bz2.open
            freader = PysReader
        elif filepath.suffix == ".xlsx":
            if openpyxl is None:
                msg = f"openpyxl is not installed. {filepath} not opened."
                self.main_window.statusBar().showMessage(msg)
                return
            fopen = open
            freader = XlsxReader
        else:
            msg = f"Unknown file format {filepath.suffix}. "\
                  f"{filepath} not opened."
            self.main_window.statusBar().showMessage(msg)
            return

        # Process events before showing the modal progress dialog
        QApplication.instance().processEvents()

        # Change into file directory
        os.chdir(filepath.parent)

        # Load file into grid
        title = "File open progress"
        label = f"Opening {filepath.name}..."

        try:
            with fopen(filepath, "rb") as infile:
                reader = freader(infile, code_array)
                try:
                    for i, _ in file_progress_gen(self.main_window, reader,
                                                  title, label, filelines):
                        pass
                except Exception as error:
                    logging.error(error)
                    grid.model.reset()
                    self.main_window.statusBar().showMessage(str(error))
                    self.main_window.safe_mode = False
                    return
                except ProgressDialogCanceled:
                    msg = f"File open stopped by user at line {i}."
                    logging.info(msg)
                    self.main_window.statusBar().showMessage(msg)
                    grid.model.reset()
                    self.main_window.safe_mode = False
                    return

        except Exception as err:
            # A lot may got wrong with a malformed pys file, includes OSError
            msg_tpl = "Error opening file {filepath}: {err}."
            msg = msg_tpl.format(filepath=filepath, err=err)
            self.main_window.statusBar().showMessage(msg)
            # Reset grid
            grid.model.reset()
            self.main_window.safe_mode = False
            return
        # Explicitly set the grid shape
        shape = code_array.shape
        grid.model.shape = shape

        # Update cell spans and zoom because this is unsupported by the model
        for grid in self.main_window.grids:
            with grid.undo_resizing_row():
                with grid.undo_resizing_column():
                    grid.update_cell_spans()
                    grid.update_zoom()

            # Update index widgets
            grid.update_index_widgets()

            grid.model.dataChanged.emit(QModelIndex(), QModelIndex())

            # Select upper left cell because initial selection oddities
            grid.reset_selection()

        # Change the main window last input directory state
        self.main_window.settings.last_file_input_path = filepath
        self.main_window.settings.last_file_output_path = filepath

        # Change the main window filepath state
        self.main_window.settings.changed_since_save = False

        # Update macro editor
        self.main_window.macro_panel.update_()

        # Add to file history
        self.main_window.settings.add_to_file_history(filepath.as_posix())

        # Update recent files in the file menu
        self.main_window.menuBar().file_menu.history_submenu.update_()

        return filepath

    @handle_changed_since_save
    def file_open(self):
        """File open workflow"""

        # Get filepath from user
        dial = FileOpenDialog(self.main_window)
        if not dial.file_path:
            return  # Cancel pressed
        filepath = Path(dial.file_path).with_suffix(dial.suffix)

        self.filepath_open(filepath)

    @handle_changed_since_save
    def file_open_recent(self, filepath: Path):
        """File open recent workflow

        :param filepath: Path of file to be opened

        """

        self.filepath_open(Path(filepath))

    def sign_file(self, filepath: Path):
        """Signs filepath if not in :attr:`model.model.DataArray.safe_mode`

        :param filepath: Path of file to be signed

        """

        if self.main_window.safe_mode:
            msg = "File saved but not signed because it is unapproved."
            self.main_window.statusBar().showMessage(msg)
            return

        signature_key = self.main_window.settings.signature_key
        try:
            with open(filepath, "rb") as infile:
                signature = sign(infile.read(), signature_key)
        except OSError as err:
            msg = f"Error signing file: {err}"
            self.main_window.statusBar().showMessage(msg)
            return

        if signature is None or not signature:
            msg = 'Error signing file.'
            self.main_window.statusBar().showMessage(msg)
            return

        signature_path = filepath.with_suffix(filepath.suffix + '.sig')
        try:
            with open(signature_path, 'wb') as signfile:
                signfile.write(signature)
                msg = "File saved and signed."
        except OSError as err:
            msg_tpl = "Error signing file {filepath}: {err}."
            msg = msg_tpl.format(filepath=filepath, err=err)

        self.main_window.statusBar().showMessage(msg)

    def _save(self, filepath: Path):
        """Save filepath using chosen_filter

        Compresses save file if filepath.suffix is `.pys`

        :param filepath: Path of file to be saved

        """

        code_array = self.main_window.grid.model.code_array

        # Process events before showing the modal progress dialog
        QApplication.instance().processEvents()

        # Save grid to temporary file

        title = "File save progress"
        label = f"Saving {filepath.name}..."

        with NamedTemporaryFile(delete=False) as tempfile:
            filename = tempfile.name
            try:
                pys_writer = PysWriter(code_array)
                try:
                    for _, line in file_progress_gen(
                            self.main_window, pys_writer, title, label,
                            len(pys_writer)):
                        line = bytes(line, "utf-8")
                        if filepath.suffix == ".pys":
                            line = bz2.compress(line)
                        tempfile.write(line)

                except ProgressDialogCanceled:
                    msg = "File save stopped by user."
                    self.main_window.statusBar().showMessage(msg)
                    tempfile.delete = True  # Delete incomplete tmpfile
                    return False

            except (OSError, ValueError) as err:
                tempfile.delete = True
                QMessageBox.critical(self.main_window, "Error saving file",
                                     str(err))
                return False
        try:
            if filepath.exists() and not os.access(filepath, os.W_OK):
                raise PermissionError(f"No write access to {filepath}")
            move(filename, filepath)

        except OSError as err:
            # No tmp file present
            QMessageBox.critical(self.main_window, "Error saving file",
                                 str(err))
            return False

        # Change the main window filepath state
        self.main_window.settings.changed_since_save = False

        # Set the current filepath
        self.main_window.settings.last_file_output_path = filepath

        # Change the main window title
        self.main_window.setWindowTitle(f"{filepath.name} - pyspread")

        # Add to file history
        self.main_window.settings.add_to_file_history(filepath.as_posix())

        # Update recent files in the file menu
        self.main_window.menuBar().file_menu.history_submenu.update_()

        self.sign_file(filepath)

    def file_save(self):
        """File save workflow"""

        filepath = self.main_window.settings.last_file_output_path
        if filepath.suffix and self._save(filepath) is not False:
            return

        # New, changed file that has never been saved before
        # Now the user has aborted the file save as dialog or
        # there was a write error
        return self.file_save_as()

    def file_save_as(self):
        """File save as workflow"""

        # Get filepath from user
        dial = FileSaveDialog(self.main_window)
        if not dial.file_path:
            return False  # Cancel pressed

        filepath = Path(dial.file_path)

        # Extend filepath suffix if needed
        if filepath.suffix != dial.suffix:
            filepath = filepath.with_suffix(dial.suffix)

        return self._save(filepath)

    def file_import(self):
        """Import csv files"""

        # Get filepath from user
        dial = CsvFileImportDialog(self.main_window)
        if not dial.file_path:
            return  # Cancel pressed
        filepath = Path(dial.file_path)

        self._csv_import(filepath)

    def _csv_import(self, filepath: Path):
        """Import csv from filepath

        :param filepath: Path of file to be imported

        """

        filelines = self.count_file_lines(filepath)
        if not filelines:  # May not be None or 0
            title = "CSV Import Error"
            text = f"File {filepath} seems to be empty."
            QMessageBox.warning(self.main_window, title, text)
            return

        # Store file import path for next time importing a file
        self.main_window.settings.last_file_import_path = filepath

        digest_types = self.main_window.settings.digest_types

        csv_dlg = CsvImportDialog(self.main_window, filepath,
                                  digest_types=digest_types)

        if not csv_dlg.exec():
            return

        self.main_window.settings.digest_types = csv_dlg.digest_types
        dialect = csv_dlg.dialect
        digest_types = csv_dlg.digest_types
        try:
            keep_header = dialect.hasheader and dialect.keepheader
        except AttributeError:
            keep_header = False
        grid = self.main_window.focused_grid
        row, column, _ = current = grid.current
        model = grid.model
        rows, columns, tables = model.shape

        # Dialog accepted, now check if grid is large enough
        csv_rows = filelines
        if dialect.hasheader and not dialect.keepheader:
            csv_rows -= 1
        csv_columns = csv_dlg.csv_table.model.columnCount()
        max_rows, max_columns = self.main_window.settings.maxshape[:2]

        if csv_rows > rows - row or csv_columns > columns - column:
            if csv_rows + row > max_rows or csv_columns + column > max_columns:
                # Required grid size is too large
                text_tpl = "The csv file {} does not fit into the grid.\n " +\
                           "\nIt has {} rows and {} columns. Counting from " +\
                           "the current cell, {} rows and {} columns would " +\
                           "be needed, which exeeds the maximum shape of " +\
                           "{} rows and {} columns. Data that does not fit " +\
                           "inside the grid is discarded.\n \nDo you want " +\
                           "to increase the grid size so that as much data " +\
                           "from the csv file as possible fits in?"
                text = text_tpl.format(filepath, csv_rows, csv_columns,
                                       rows-row, columns-column, max_rows,
                                       max_columns)
            else:
                # Shall we resize the grid?
                text_tpl = \
                    "The csv file {} does not fit into the grid.\n \n" +\
                    "It has {} rows and {} columns. Counting from the " +\
                    "current cell, only {} rows and {} columns remain for " +\
                    "CSV data.\n \nData that does not fit inside the grid " +\
                    "is discarded.\n \nDo you want to increase the grid " +\
                    "size so that all csv file data fits in?"
                text = text_tpl.format(filepath, csv_rows, csv_columns,
                                       rows-row, columns-column)

            title = "CSV Content Exceeds Grid Shape"
            choices = QMessageBox.StandardButton.No \
                | QMessageBox.StandardButton.Yes \
                | QMessageBox.StandardButton.Cancel
            default_choice = QMessageBox.StandardButton.No
            choice = QMessageBox.question(self.main_window, title, text,
                                          choices, default_choice)
            if choice == QMessageBox.StandardButton.Yes:
                # Resize grid
                target_rows = min(max_rows, max(csv_rows + row, rows))
                target_columns = min(max_columns,
                                     max(csv_columns + column, columns))
                self._resize_grid((target_rows, target_columns, tables))
                rows = target_rows
                columns = target_columns

            elif choice == QMessageBox.StandardButton.Cancel:
                return

        # Now fill the grid

        description_tpl = "Import from csv file {} at cell {}"
        description = description_tpl.format(filepath, current)

        command = None

        title = "csv import progress"
        label = f"Importing {filepath.name}..."

        try:
            if hasattr(dialect, "encoding"):
                __encoding = dialect.encoding
            else:
                __encoding = csv_dlg.csv_encoding
            with open(filepath, newline='', encoding=__encoding) as csvfile:
                try:
                    reader = csv_reader(csvfile, dialect)
                    for i, line in file_progress_gen(self.main_window, reader,
                                                     title, label, filelines):
                        if row + i >= rows:
                            break

                        for j, ele in enumerate(line):
                            if column + j >= columns:
                                break

                            if digest_types is None:
                                code = str(ele)
                            elif i == 0 and keep_header:
                                code = repr(ele)
                            else:
                                code = convert(ele, digest_types[j])
                            index = model.index(row + i, column + j)
                            _command = commands.SetCellCode(code, model, index,
                                                            description)
                            try:
                                command.mergeWith(_command)
                            except AttributeError:
                                command = _command

                except (TypeError, ValueError) as error:
                    title = "CSV Import Error"
                    text_tpl = "Error importing csv file {path}.\n \n" + \
                               "{errtype}: {error}"
                    text = text_tpl.format(path=filepath,
                                           errtype=type(error).__name__,
                                           error=error)
                    QMessageBox.warning(self.main_window, title, text)
                    return

                except ProgressDialogCanceled:
                    title = "CSV Import Stopped"
                    text = f"Import stopped by user at line {i}."
                    QMessageBox.warning(self.main_window, title, text)
                    return

        except Exception as error:
            # A lot may go wrong with malformed csv files, includes OSError
            title = "CSV Import Error"
            text_tpl = "Error importing csv file {path}.\n \n" +\
                       "{errtype}: {error}"
            text = text_tpl.format(path=filepath, errtype=type(error).__name__,
                                   error=error)
            QMessageBox.warning(self.main_window, title, text)
            return

        with self.main_window.entry_line.disable_updates():
            with self.busy_cursor():
                with self.prevent_updates():
                    if command is not None:
                        self.main_window.undo_stack.push(command)

    def file_export(self):
        """Export csv and svg files"""

        # Determine what filters ae available
        filters_list = ["CSV (*.csv)", "SVG (*.svg)"]

        grid = self.main_window.focused_grid

        current = grid.current
        code_array = grid.model.code_array

        res = code_array[current]

        if isinstance(res, QImage):
            filters_list.append("JPG of current cell (*.jpg)")

        if isinstance(res, (QImage, matplotlib.figure.Figure)):
            filters_list.append("PNG of current cell (*.png)")

        if isinstance(res, matplotlib.figure.Figure):
            filters_list.append("SVG of current cell (*.svg)")

        # Get filepath from user
        dial = FileExportDialog(self.main_window, filters_list)
        if not dial.file_path:
            return  # Cancel pressed
        filepath = Path(dial.file_path)

        # Store file export path for next time exporting a file
        self.main_window.settings.last_file_export_path = filepath

        if "CSV" in dial.selected_filter:
            self._csv_export(filepath)
            return

        if "SVG" in dial.selected_filter:
            # Extend filepath suffix if needed
            if filepath.suffix != dial.suffix:
                filepath = filepath.with_suffix(dial.suffix)
            self.svg_export(filepath)
            return

        # Extend filepath suffix if needed
        if filepath.suffix != dial.suffix:
            filepath = filepath.with_suffix(dial.suffix)

        if "JPG" in dial.selected_filter:
            if isinstance(res, QImage):
                self._qimage_export(str(filepath), file_format="jpg")

        if "PNG" in dial.selected_filter:
            if isinstance(res, QImage):
                self._qimage_export(str(filepath), file_format="png")
            elif isinstance(res, matplotlib.figure.Figure):
                self._matplotlib_export(filepath, file_format="png")

        elif "SVG" in dial.selected_filter:
            if isinstance(res, matplotlib.figure.Figure):
                self._matplotlib_export(filepath, file_format="svg")

    def _csv_export(self, filepath: Path):
        """Export to csv file filepath

        :param filepath: Path of file to be exported

        """

        grid = self.main_window.focused_grid

        # Get area for csv export
        area = CsvExportAreaDialog(self.main_window, grid,
                                   title="Csv export area").area
        if area is None:
            return

        code_array = grid.model.code_array
        table = grid.table
        csv_data = code_array[area.top: area.bottom + 1,
                              area.left: area.right + 1, table]

        csv_dlg = CsvExportDialog(self.main_window, area)

        if not csv_dlg.exec():
            return

        try:
            with open(filepath, "w", newline='', encoding='utf-8') as csvfile:
                writer = csv.writer(csvfile, dialect=csv_dlg.dialect)
                writer.writerows(csv_data)
        except OSError as error:
            self.main_window.statusBar().showMessage(str(error))

    def svg_export(self, filepath: Path, svg_area: SinglePageArea = None):
        """Export to svg file filepath

        :param filepath: Path of file to be exported
        :param svg_area: Area of the grid to be exported

        """

        with self.print_zoom():
            grid = self.main_window.grid

            generator = QSvgGenerator()
            generator.setFileName(str(filepath))

            if svg_area is None:
                # Get area for svg export
                svg_area = SvgExportAreaDialog(self.main_window, grid,
                                               title="Svg export area").area
            if svg_area is None:
                return

            self.main_window.print_area = svg_area

            rows = self.get_paint_rows(svg_area.top, svg_area.bottom)
            columns = self.get_paint_columns(svg_area.left, svg_area.right)
            total_height = self.get_total_height(svg_area.top, svg_area.bottom)
            total_width = self.get_total_width(svg_area.left, svg_area.right)

            x_offset = grid.columnViewportPosition(0)
            y_offset = grid.rowViewportPosition(0)

            top_left_idx = grid.model.index(svg_area.top, svg_area.left)
            top_left_visual_rect = grid.visualRect(top_left_idx)

            generator.setSize(QSize(total_width, total_height))
            paint_rect = QRectF(top_left_visual_rect.x() - x_offset,
                                top_left_visual_rect.y() - y_offset,
                                total_width,
                                total_height)
            generator.setViewBox(paint_rect)
            option = QStyleOptionViewItem()

            painter = QPainter(generator)
            self.paint(painter, option, paint_rect, rows, columns)

            painter.end()
            self.main_window.print_area = None

    def _qimage_export(self, filepath: Path, file_format: str):
        """Export to png file filepath

        :param filepath: Path of file to be exported
        :param file_format: File format to be exported, e.g. png

        """

        grid = self.main_window.focused_grid
        code_array = grid.model.code_array
        qimage = code_array[grid.current]

        try:
            if not qimage.save(filepath, file_format):
                msg = f"Could not save {filepath}"
                self.main_window.statusBar().showMessage(msg)
        except Exception as error:
            self.main_window.statusBar().showMessage(str(error))

    def _matplotlib_export(self, filepath: Path, file_format: str):
        """Export to svg file filepath

        :param filepath: Path of file to be exported
        :param file_format: File format to be exported, e.g. png or svg

        """

        if matplotlib is None:
            # matplotlib is not installed
            return

        grid = self.main_window.focused_grid
        code_array = grid.model.code_array
        figure = code_array[grid.current]

        try:
            figure.savefig(filepath, format=file_format)
        except Exception as error:
            self.main_window.statusBar().showMessage(str(error))

    @contextmanager
    def print_zoom(self, zoom: float = 1.0):
        """Decorator for tasks that have to take place in standard zoom

        :param zoom: Print zoom factor

        """

        grid = self.main_window.focused_grid
        __zoom = grid.zoom
        grid.zoom = zoom
        yield
        grid.zoom = __zoom

    def get_paint_rows(self, top: int, bottom: int) -> Iterable[int]:
        """Iterator of rows to paint

        :param top: First row to paint
        :param bottom: Last row to paint

        """

        grid = self.main_window.focused_grid
        rows = grid.model.shape[0]
        top = max(0, min(rows - 1, top))
        bottom = max(0, min(rows - 1, bottom))
        if top == -1:
            top = 0
        if bottom == -1:
            bottom = grid.model.shape[0]

        return range(top, bottom + 1)

    def get_paint_columns(self, left: int, right: int) -> Iterable[int]:
        """Iterator of columns to paint

        :param left: First column to paint
        :param right: Last column to paint

        """

        grid = self.main_window.focused_grid
        columns = grid.model.shape[1]
        left = max(0, min(columns - 1, left))
        right = max(0, min(columns - 1, right))
        if left == -1:
            left = 0
        if right == -1:
            right = grid.model.shape[1]

        return range(left, right + 1)

    def get_paint_tables(self, first: int, last: int) -> Iterable[int]:
        """Iterator of tables to paint

        :param first: First table to paint
        :param last: Last table to paint

        """

        grid = self.main_window.focused_grid
        tables = grid.model.shape[2]
        first = max(0, min(tables - 1, first))
        last = max(0, min(tables - 1, last))
        if first == -1:
            first = 0
        if last == -1:
            last = grid.model.shape[2]

        return range(first, last + 1)

    def get_total_height(self, top: int, bottom: int) -> float:
        """Total height of paint_rows

        :param top: First row to evaluate
        :param bottom: Last row to evaluate

        """

        grid = self.main_window.focused_grid
        rows = self.get_paint_rows(top, bottom)
        return sum(grid.rowHeight(row) for row in rows)

    def get_total_width(self, left: int, right: int) -> float:
        """Total height of paint_columns

        :param left: First column to evaluate
        :param right: Last column to evaluate

        """

        grid = self.main_window.focused_grid
        columns = self.get_paint_columns(left, right)
        return sum(grid.columnWidth(column) for column in columns)

    def paint(self, painter: QPainter, option: QStyleOptionViewItem,
              paint_rect: QRectF, rows: Iterable[int], columns: Iterable[int]):
        """Grid paint workflow for printing and svg export

        :param painter: Painter with which the grid is drawn
        :param option: Style option for rendering
        :param paint_rect: Rectangle, which is drawn at the grid borders
        :param rows: Rows to be painted
        :param columns: Columns to be painted

        """

        grid = self.main_window.focused_grid
        code_array = grid.model.code_array
        cell_attributes = code_array.cell_attributes

        x_offset = grid.columnViewportPosition(0)
        y_offset = grid.rowViewportPosition(0)

        max_width = 0
        max_height = 0

        for row in rows:
            for column in columns:
                key = row, column, grid.table
                merging_cell = cell_attributes.get_merging_cell(key)
                if merging_cell is None \
                   or merging_cell[0] == row and merging_cell[1] == column:

                    idx = grid.model.index(row, column)
                    visual_rect = grid.visualRect(idx)
                    x = max(0, visual_rect.x() - x_offset)
                    y = max(0, visual_rect.y() - y_offset)
                    width = visual_rect.width()
                    if visual_rect.x() - x_offset < 0:
                        width += visual_rect.x() - x_offset
                    height = visual_rect.height()
                    if visual_rect.y() - y_offset < 0:
                        height += visual_rect.y() - y_offset

                    option.rect = QRect(int(x), int(y),
                                        int(width), int(height))
                    option.rectf = QRectF(x, y, width, height)

                    max_width = max(max_width, x + width)
                    max_height = max(max_height, y + height)
                    # painter.setClipRect(option.rectf)

                    option.text = code_array(key)
                    option.widget = grid

                    grid.itemDelegate().paint(painter, option, idx)

    @handle_changed_since_save
    def file_quit(self):
        """Program exit workflow"""

        self.main_window.settings.save()
        QApplication.instance().quit()

    # Edit menu

    def delete(self, description_tpl: str = "Delete selection {}",
               selection: Selection = None):
        """Delete cells in selection

        :param description_tpl: Description template for `QUndoCommand`
        :param selection: Selection to delete, deletes grid selection if None

        """

        grid = self.main_window.focused_grid
        model = grid.model
        if selection is None:
            selection = grid.selection

        description = description_tpl.format(selection)
        command = commands.DeleteSelectedCellData(grid, model, selection,
                                                  description)
        self.main_window.undo_stack.push(command)

    def edit_cut(self, *, description_tpl: str = "Cut selection {}",
                 selection: Selection = None):
        """Edit -> Cut workflow

        :param description_tpl: Description template for `QUndoCommand`
        :param selection: Selection to cut, cuts grid selection if None

        """

        self.edit_copy(selection=selection)
        self.delete(description_tpl=description_tpl, selection=selection)

    def edit_copy(self, *, selection: Selection = None):
        """Edit -> Copy workflow

        Copies selected grid code to clipboard

        :param selection: Selection to copy, copies grid selection if None

        """

        grid = self.main_window.focused_grid
        table = grid.table
        if selection is None:
            selection = grid.selection
        bbox = selection.get_grid_bbox(grid.model.shape)
        (top, left), (bottom, right) = bbox

        data = []

        for row in range(top, bottom + 1):
            data.append([])
            for column in range(left, right + 1):
                if (row, column) in selection:
                    try:
                        code = grid.model.code_array((row, column, table))
                    except IndexError:
                        code = None
                    if code is None:
                        code = ""
                    code = code.replace("\n", "\u000C")  # Replace LF by FF
                else:
                    code = ""
                data[-1].append(code)

        data_string = "\n".join("\t".join(line) for line in data)

        clipboard = QApplication.clipboard()
        clipboard.setText(data_string)

    def _copy_results_current(self, grid: QTableView):
        """Copy cell results for the current cell

        :param grid: Main grid

        """

        current = grid.current
        data = grid.model.code_array[current]
        if data is None:
            return

        clipboard = QApplication.clipboard()

        # Get renderer for current cell
        renderer = grid.model.code_array.cell_attributes[current].renderer

        if renderer == "text":
            clipboard.setText(grid.model.data(grid.currentIndex()))
            # clipboard.setText(str(data))

        elif renderer == "image":
            if isinstance(data, QImage):
                clipboard.setImage(data)
            else:
                # We may have an svg image here
                try:
                    svg_bytes = bytes(data)
                except TypeError:
                    svg_bytes = bytes(data, encoding='utf-8')
                if is_svg(svg_bytes):
                    mime_data = QMimeData()
                    mime_data.setData("image/svg+xml", svg_bytes)
                    clipboard.setMimeData(mime_data)

        elif renderer == "markup":
            mime_data = QMimeData()
            mime_data.setHtml(str(data))

            # Also copy data as plain text
            doc = QTextDocument()
            doc.setHtml(str(data))
            mime_data.setText(doc.toPlainText())

            clipboard.setMimeData(mime_data)

        elif renderer == "matplotlib" and isinstance(data,
                                                     matplotlib_figure.Figure):
            # We copy and svg to the clipboard
            svg_filelike = io.BytesIO()
            png_filelike = io.BytesIO()
            data.savefig(svg_filelike, format="svg")
            data.savefig(png_filelike, format="png")
            svg_bytes = (svg_filelike.getvalue())
            png_image = QImage().fromData(png_filelike.getvalue())
            mime_data = QMimeData()
            mime_data.setData("image/svg+xml", svg_bytes)
            mime_data.setImageData(png_image)
            clipboard.setMimeData(mime_data)

    def _copy_results_selection(self, grid: QTableView):
        """Copy selected cells result objects to the clipboard

        :param grid: Main grid

        """

        def str_nn(ele):
            """str which returns '' if ele is None"""
            if ele is None:
                return ''

            try:
                if ele.__class__ in class_format_functions:
                    format_function = class_format_functions[ele.__class__]
                    return format_function(ele)

                return str(ele)
            except Exception as err:
                return str(err)

        table = grid.table
        selection = grid.selection
        bbox = selection.get_grid_bbox(grid.model.shape)
        (top, left), (bottom, right) = bbox

        data = grid.model.code_array[top:bottom+1, left:right+1, table]
        data_string = "\n".join("\t".join(map(str_nn, line)) for line in data)

        clipboard = QApplication.clipboard()
        clipboard.setText(data_string)

    def edit_copy_results(self):
        """Edit -> Copy results workflow

        If a selection is present then repr of selected grid cells result
        objects are copied to the clipboard.

        If no selection is present, the current cell results are copied to the
        clipboard. This can be plain text, html, a png image or an svg image.

        """

        grid = self.main_window.focused_grid

        if grid.has_selection():
            self._copy_results_selection(grid)
        else:
            self._copy_results_current(grid)

    def _paste_to_selection(self, selection: Selection, data: str,
                            description_tpl: str = "Paste clipboard to {}"):
        """Pastes data into grid filling the selection

        :param selection: Grid cell selection for pasting
        :param data: Clipboard text
        :param description_tpl: Description template for `QUndoCommand`

        """

        grid = self.main_window.focused_grid
        model = grid.model
        undo_stack = self.main_window.undo_stack

        description = description_tpl.format(selection)

        cmd = commands.PasteSelectedCellData(grid, model, selection, data,
                                             description)
        undo_stack.push(cmd)

    def _paste_to_current(self, data: str):
        """Pastes data into grid starting from the current cell

        :param data: Clipboard text

        """

        grid = self.main_window.focused_grid
        model = grid.model
        top, left, table = current = grid.current
        code_array = grid.model.code_array
        undo_stack = self.main_window.undo_stack

        description_tpl = "Paste clipboard starting from cell {}"
        description = description_tpl.format(current)

        command = None

        paste_gen = (line.split("\t") for line in data.split("\n"))
        for row, line in enumerate(paste_gen):
            paste_row = row + top
            if (paste_row, 0, table) not in code_array:
                break
            for column, value in enumerate(line):
                paste_column = column + left
                if (paste_row, paste_column, table) in code_array:
                    index = model.index(paste_row, paste_column, QModelIndex())
                    # Preserve line breaks
                    value = value.replace("\u000C", "\n")
                    cmd = commands.SetCellCode(value, model, index,
                                               description)
                    if command is None:
                        command = cmd
                    else:
                        command.mergeWith(cmd)
                else:
                    break
        if command is not None:
            undo_stack.push(command)

    def edit_paste(self):
        """Edit -> Paste workflow

        Pastes text clipboard data

        If no selection is present, data is pasted starting with the current
        cell. If a selection is present, data is pasted fully if the selection
        is smaller. If the selection is larger then data is duplicated.

        """

        grid = self.main_window.focused_grid

        clipboard = QApplication.clipboard()
        data = clipboard.text()

        if data:
            # Change the main window filepath state
            self.main_window.settings.changed_since_save = True

            with self.busy_cursor():
                if grid.has_selection():
                    self._paste_to_selection(grid.selection, data)
                else:
                    self._paste_to_current(data)

    def _paste_svg(self, svg: str, index: QModelIndex):
        """Pastes svg image into cell

        :param svg: SVG data
        :param index: Target cell index

        """

        grid = self.main_window.focused_grid

        codelines = svg.splitlines()
        codelines[0] = '"""' + codelines[0]
        codelines[-1] = codelines[-1] + '"""'
        code = "\n".join(codelines)

        model = grid.model
        description = f"Insert svg image into cell {index}"

        grid.on_image_renderer_pressed()
        with self.main_window.entry_line.disable_updates():
            command = commands.SetCellCode(code, model, index, description)
            self.main_window.undo_stack.push(command)

    def _paste_image(self, image_data: bytes, index: QModelIndex):
        """Pastes svg image into cell

        :param image_data: Raw image data. May be anything that QImage handles.
        :param index: Target cell index

        """

        def gen_chunk(string: str, length: int) -> Iterable[str]:
            """Generator for chunks of string

            :param string: String to be chunked
            :param length: Chunk length

            """

            for i in range(0, len(string), length):
                yield string[i:i+length]

        grid = self.main_window.focused_grid

        repr_image_data = repr(b85encode(bz2.compress(image_data)))
        newline = "'\n+b'"

        image_data_str = newline.join(gen_chunk(repr_image_data, 8000))

        code_lines = [
            "data = bz2.decompress(base64.b85decode(",
            image_data_str,
            "))",
            "qimg = QImage()",
            "QImage.loadFromData(qimg, data)",
            "qimg",
        ]

        code = "\n".join(code_lines)

        model = grid.model
        description = f"Insert image into cell {index}"

        grid.on_image_renderer_pressed()
        with self.main_window.entry_line.disable_updates():
            command = commands.SetCellCode(code, model, index, description)
            self.main_window.undo_stack.push(command)

    def edit_paste_as(self):
        """Pastes clipboard into one cell using a user specified mime type"""

        grid = self.main_window.focused_grid
        model = grid.model

        # The mimetypes that are supported by pyspread
        mimetypes = ("image", "text/html", "text/plain")
        clipboard = QApplication.clipboard()
        formats = clipboard.mimeData().formats()

        items = [fmt for fmt in formats if any(m in fmt for m in mimetypes)]
        if not items:
            return
        if len(items) == 1:
            item = items[0]
        else:
            item, ok = QInputDialog.getItem(self.main_window, "Paste as",
                                            "Choose mime type", items,
                                            current=0, editable=False)
            if not ok:
                return

        row, column, _ = current = grid.current  # Target cell key

        description_tpl = "Paste {} from clipboard into cell {}"
        description = description_tpl.format(item, current)

        index = model.index(row, column, QModelIndex())

        mime_data = clipboard.mimeData()

        if item == "image/svg+xml":
            # SVG Image
            if mime_data:
                svg = mime_data.data("image/svg+xml")
                self._paste_svg(str(svg, encoding='utf-8'), index)

        elif "image" in item and mime_data.hasImage():
            # Bitmap Image
            image = clipboard.image()
            buffer = QBuffer()
            buffer.open(QBuffer.OpenModeFlag.ReadWrite)
            image.save(buffer, "PNG")
            buffer.seek(0)
            image_data = buffer.readAll()
            buffer.close()
            self._paste_image(image_data, index)

        elif item == "text/html" and mime_data.hasHtml():
            # HTML content
            html = mime_data.html()
            command = commands.SetCellCode(html, model, index, description)
            self.main_window.undo_stack.push(command)
            grid.on_markup_renderer_pressed()

        elif item == "text/plain":
            # Normal code
            code = clipboard.text()
            if code:
                command = commands.SetCellCode(code, model, index, description)
                self.main_window.undo_stack.push(command)

        else:
            # Unknown mime type
            return NotImplemented

    def edit_find(self):
        """Edit -> Find workflow, opens FindDialog"""

        find_dialog = FindDialog(self.main_window)
        find_dialog.show()
        find_dialog.raise_()
        find_dialog.activateWindow()

    def _get_next_match(self, find_dialog: FindDialog,
                        start_key: Tuple[int, int, int] = None):
        """Returns tuple of find string and next matching cell key

        :param find_dialog: Find dialog from which the search origins
        :param start_key: Start search at given key

        """

        grid = self.main_window.focused_grid
        findnextmatch = grid.model.code_array.findnextmatch

        find_editor = find_dialog.search_text_editor
        find_string = find_editor.text()

        if start_key is not None:
            pass
        elif find_dialog.from_start_checkbox.isChecked():
            start_key = 0, 0, grid.table
        elif find_dialog.backward_checkbox.isChecked():
            start_key = grid.row - 1, grid.column, grid.table
        else:
            start_key = grid.row + 1, grid.column, grid.table

        match = findnextmatch(start_key, find_string,
                              up=find_dialog.backward_checkbox.isChecked(),
                              word=find_dialog.word_checkbox.isChecked(),
                              case=find_dialog.case_checkbox.isChecked(),
                              regexp=find_dialog.regex_checkbox.isChecked(),
                              results=find_dialog.results_checkbox.isChecked())

        return find_string, match

    def _display_match_msg(self, find_string: str, next_match: str,
                           regexp: str):
        """Displays find match message in statusbar

        :param find_string: Message component
        :param next_match: Message component
        :param regexp: Message component

        """

        str_name = "Regular expression" if regexp else "String"
        msg_tpl = "{str_name} {find_string} found in cell {next_match}."
        msg = msg_tpl.format(str_name=str_name,
                             find_string=find_string,
                             next_match=next_match)
        self.main_window.statusBar().showMessage(msg)

    def find_dialog_on_find(self, find_dialog: FindDialog):
        """Edit -> Find workflow, after pressing find button in FindDialog

        :param find_dialog: Find dialog of origin

        """

        grid = self.main_window.focused_grid

        find_string, next_match = self._get_next_match(find_dialog)

        if next_match:
            grid.current = next_match

            regexp = find_dialog.regex_checkbox.isChecked()
            self._display_match_msg(find_string, next_match, regexp)

            if find_dialog.from_start_checkbox.isChecked():
                find_dialog.from_start_checkbox.setChecked(False)

    def edit_find_next(self):
        """Edit -> Find next workflow"""

        grid = self.main_window.focused_grid

        findnextmatch = grid.model.code_array.findnextmatch

        find_editor = self.main_window.find_toolbar.find_editor
        find_string = find_editor.text()

        if find_editor.up:
            start_key = grid.row - 1, grid.column, grid.table
        else:
            start_key = grid.row + 1, grid.column, grid.table

        next_match = findnextmatch(start_key, find_string,
                                   up=find_editor.up,
                                   word=find_editor.word,
                                   case=find_editor.case,
                                   regexp=find_editor.regexp,
                                   results=find_editor.results)
        if next_match:
            grid.current = next_match

            self._display_match_msg(find_string, next_match,
                                    find_editor.regexp)

    def edit_replace(self):
        """Edit -> Replace workflow, opens ReplaceDialog"""

        find_dialog = ReplaceDialog(self.main_window)
        find_dialog.show()
        find_dialog.raise_()
        find_dialog.activateWindow()

    def _get_replace_command(self, next_match: Tuple[int, int, int],
                             find_string: str,
                             replace_string: str,
                             max_: int = 1,
                             description: str = None) -> QUndoCommand:
        """Returns SetCellCode command for replace operations

        :param next_match: Key of next matching cell
        :param find_string: String to find
        :param replace_string: Replacement string
        :param max_: Maximum number of replace actions, -1 is unlimited
        :param description: Forced command description string

        """

        model = self.main_window.focused_grid.model

        old_code = model.code_array(next_match)
        new_code = old_code.replace(find_string, replace_string, max_)

        if description is None:
            description_tpl = "Replaced {old} with {new} in cell {key}."
            description = description_tpl.format(old=old_code, new=new_code,
                                                 key=next_match)
        index = model.index(*next_match[:2])
        return commands.SetCellCode(new_code, model, index, description)

    def replace_dialog_on_replace(self, replace_dialog: ReplaceDialog,
                                  toggled: bool = False,
                                  max_: int = 1) -> bool:
        """Edit -> Replace workflow when pushing Replace in ReplaceDialog

        Returns True if there is a match otherwise False

        :param replace_dialog: Replace dialog of origin
        :param toggled: Replace dialog toggle state
        :param max_: Maximum number of replace actions, -1 is unlimited

        """

        grid = self.main_window.focused_grid

        find_string, next_match = self._get_next_match(replace_dialog)
        replace_string = replace_dialog.replace_text_editor.text()

        if next_match:
            command = self._get_replace_command(next_match, find_string,
                                                replace_string, max_=max_)
            self.main_window.undo_stack.push(command)

            grid.current = next_match

            self.main_window.statusBar().showMessage(command.description)

            if replace_dialog.from_start_checkbox.isChecked():
                replace_dialog.from_start_checkbox.setChecked(False)

            return True

        return False

    def replace_dialog_on_replace_all(self, replace_dialog: ReplaceDialog):
        """Edit -> Replace workflow when pushing ReplaceAll in ReplaceDialog

        :param replace_dialog: Replace dialog of origin

        """

        find_string = replace_dialog.search_text_editor.text()
        replace_string = replace_dialog.replace_text_editor.text()

        word = replace_dialog.word_checkbox.isChecked()
        case = replace_dialog.case_checkbox.isChecked()
        regexp = replace_dialog.regex_checkbox.isChecked()

        command = None

        grid = self.main_window.focused_grid
        code_array = grid.model.code_array
        keys = code_array.keys()

        matches = []

        with self.busy_cursor():
            with self.main_window.entry_line.disable_updates():
                with self.prevent_updates():
                    for key in keys:
                        code = code_array(key)

                        if code_array.string_match(code, find_string, word,
                                                   case, regexp) is not None:
                            matches.append(key)

                    for match in matches:
                        msg = f"Replace all {find_string} by {replace_string}"
                        _command = self._get_replace_command(match,
                                                             find_string,
                                                             replace_string,
                                                             max_=-1,
                                                             description=msg)
                        if command is None:
                            command = _command
                        else:
                            command.mergeWith(_command)

                    if command is not None:
                        self.main_window.undo_stack.push(command)

        msg = f"{find_string} replaced by {replace_string} in {len(matches)} "\
              f"cell{'s' if len(matches) != 1 else ''}."
        self.main_window.statusBar().showMessage(msg)

    def _sort(self, ascending: bool = True):
        """Edit -> Sort ascending

        :param ascending: True for ascending sort, False for descending sort

        """

        grid = self.main_window.focused_grid
        model = grid.model
        table = grid.current[2]
        selection = grid.selection

        (top, left), (bottom, right) = selection.get_bbox()
        if top == bottom:
            return

        data = grid.model.code_array[top:bottom+1, left:right+1, table].copy()
        if ascending:
            data[data == None] = numpy.inf  # `is` does not work here
        else:
            data[data == None] = -numpy.inf  # `is` does not work here

        try:
            if ascending:
                sorted_idx = data[:, grid.current[1]-left].argsort()
            else:
                sorted_idx = data[:, grid.current[1]-left].argsort()[::-1]
        except TypeError as err:
            msg = f"Could not sort selection: {err}"
            self.main_window.statusBar().showMessage(msg)
            return

        old_code = {}
        for key in selection.cell_generator(model.shape, table):
            old_code[key] = grid.model.code_array(key)

        command = None
        if ascending:
            description = f"Sort {grid.selection} ascending"
        else:
            description = f"Sort {grid.selection} descending"

        for row, column in selection.cell_generator(model.shape):
            code = old_code[(sorted_idx[row-top]+top, column, table)]
            index = model.index(row, column)
            _command = commands.SetCellCode(code, model, index, description)
            if command is None:
                command = _command
            else:
                command.mergeWith(_command)

        if command is not None:
            self.main_window.undo_stack.push(command)

        msg = "Selection sorted."
        self.main_window.statusBar().showMessage(msg)

    def edit_sort_ascending(self):
        """Edit -> Sort ascending"""

        self._sort()

    def edit_sort_descending(self):
        """Edit -> Sort descending"""

        self._sort(ascending=False)

    def edit_resize(self):
        """Edit -> Resize workflow"""

        grid = self.main_window.focused_grid

        # Get grid shape from user
        old_shape = grid.model.code_array.shape
        title = "Resize grid"
        shape = GridShapeDialog(self.main_window, old_shape, title=title).shape
        self._resize_grid(shape)

    def _resize_grid(self, shape: Tuple[int, int, int]):
        """Resize grid

        :param shape: New grid shape

        """

        grid = self.main_window.focused_grid

        old_shape = grid.model.code_array.shape

        # Check if shape is valid
        try:
            check_shape_validity(shape, self.main_window.settings.maxshape)
        except ValueError as err:
            self.main_window.statusBar().showMessage('Error: ' + str(err))
            return

        grid.current = 0, 0, 0

        description = f"Resize grid to {shape}"

        with self.main_window.entry_line.disable_updates():
            command = commands.SetGridSize(grid, old_shape, shape, description)
            self.main_window.undo_stack.push(command)

        # Select upper left cell because initial selection behaves strangely
        grid.reset_selection()

    # View menu

    def view_goto_cell(self):
        """View -> Go to cell workflow"""

        grid = self.main_window.focused_grid

        # Get cell key from user
        shape = grid.model.shape
        key = CellKeyDialog(self.main_window, shape).key

        if key is not None:
            grid.current = key

    # Format menu

    def format_copy_format(self):
        """Copies the format of the selected cells to the Clipboard

        Cells are shifted so that the top left bbox corner is at 0,0

        """

        def remove_tabu_keys(attrs: AttrDict):
            """Remove keys that are not copied from attr

            :param attr: Attribute dict that holds cell attributes

            """

            tabu_attrs = "merge_area", "frozen"
            for tabu_attr in tabu_attrs:
                try:
                    attrs.pop(tabu_attr)
                except KeyError:
                    pass

        grid = self.main_window.focused_grid
        code_array = grid.model.code_array
        cell_attributes = code_array.cell_attributes

        # Cell attributes

        new_cell_attributes = []
        selection = grid.selection

        # Format content is shifted so that the top left corner is 0,0
        (top, left), (_, _) = selection.get_grid_bbox(grid.model.shape)

        table_cell_attributes = cell_attributes.for_table(grid.table)
        for __selection, _, attrs in table_cell_attributes:
            new_selection = selection & __selection
            if new_selection:
                # We do not copy merged cells and cell renderers
                __attrs = copy(attrs)
                remove_tabu_keys(__attrs)
                new_shifted_selection = new_selection.shifted(-top, -left)
                if __attrs:
                    cell_attribute = new_shifted_selection.parameters, __attrs
                    new_cell_attributes.append(cell_attribute)

        ca_repr = bytes(repr(new_cell_attributes), encoding='utf-8')

        clipboard = QApplication.clipboard()
        mime_data = QMimeData()
        mime_data.setData("application/x-pyspread-cell-attributes", ca_repr)
        clipboard.setMimeData(mime_data)

    def format_paste_format(self):
        """Pastes cell formats

        Pasting starts at cursor or at top left bbox corner

        """

        clipboard = QApplication.clipboard()
        mime_data = clipboard.mimeData()

        grid = self.main_window.focused_grid
        model = grid.model

        row, column, table = grid.current

        if "application/x-pyspread-cell-attributes" not in mime_data.formats():
            return

        cas_data = mime_data.data("application/x-pyspread-cell-attributes")
        cas_data_str = str(cas_data, encoding='utf-8')
        cas = literal_eval(cas_data_str)
        if not isinstance(cas, list):
            msg_tpl = "{} has type {} that is not instance of list"
            msg = msg_tpl.format(cas, type(cas))
            raise Warning(msg)

        tabu_attrs = "merge_area", "frozen"

        description_tpl = "Paste format for selections {}"
        description = description_tpl.format([ca[0] for ca in cas])

        for selection_params, attrs in cas:
            if not any(tabu_attr in attrs for tabu_attr in tabu_attrs):
                selection = Selection(*selection_params)
                shifted_selection = selection.shifted(row, column)
                attr_dict = AttrDict()
                attr_dict.update(attrs)
                new_cell_attribute = CellAttribute(shifted_selection, table,
                                                   attr_dict)
                selected_idx = []
                for key in shifted_selection.cell_generator(model.shape):
                    selected_idx.append(model.index(*key))
                command = commands.SetCellFormat(new_cell_attribute, model,
                                                 grid.currentIndex(),
                                                 selected_idx, description)
                self.main_window.undo_stack.push(command)

    # Macro menufilepath

    def _read_svg_str(self, filepath, encoding):
        """Returns svg string from filepath

        :param filepath: Path of SVG file to read

        """

        try:
            with open(filepath, "r", encoding=encoding) as svgfile:
                return svgfile.read()
        except UnicodeError:
            encoding, ok = QInputDialog().getItem(
                self, f"{filepath} not encoded in utf-8",
                f"Encoding of {filepath}",
                self.main_window.settings.encodings)
            if ok:
                return self._read_svg_str(filepath, encoding)
        except OSError as err:
            msg_tpl = "Error opening file {filepath}: {err}."
            msg = msg_tpl.format(filepath=filepath, err=err)
            self.main_window.statusBar().showMessage(msg)

    def macro_insert_image(self):
        """Insert image workflow"""

        grid = self.main_window.focused_grid

        dial = ImageFileOpenDialog(self.main_window)
        if not dial.file_path:
            return  # Cancel pressed

        filepath = Path(dial.file_path)

        index = grid.currentIndex()
        grid.clearSelection()
        grid.selectionModel().select(index,
                                     QItemSelectionModel.SelectionFlag.Select)

        if filepath.suffix == ".svg":
            svg = self._read_svg_str(filepath, encoding='utf-8')
            if not svg:
                return

            self._paste_svg(svg, index)
        else:
            try:
                with open(filepath, "rb") as imgfile:
                    image_data = imgfile.read()
            except OSError as err:
                msg_tpl = "Error opening file {filepath}: {err}."
                msg = msg_tpl.format(filepath=filepath, err=err)
                self.main_window.statusBar().showMessage(msg)
                return
            self._paste_image(image_data, index)

    def macro_insert_chart(self):
        """Insert chart workflow"""

        grid = self.main_window.focused_grid

        model = grid.model
        code_array = model.code_array
        current = grid.current
        code = code_array(current)

        if current in self.cell2dialog:
            self.cell2dialog[current].activateWindow()
            self.cell2dialog[current].setFocus()
            return

        chart_dialog = ChartDialog(self.main_window, current)
        self.cell2dialog[current] = chart_dialog

        chart_dialog.show()

        if code is not None:
            chart_dialog.editor.setPlainText(code)

        if chart_dialog.exec() == ChartDialog.DialogCode.Accepted:
            code = chart_dialog.editor.toPlainText()
            grid.current = current
            index = grid.currentIndex()
            grid.clearSelection()
            grid.selectionModel().select(
                index, QItemSelectionModel.SelectionFlag.Select)
            grid.on_matplotlib_renderer_pressed()

            description = f"Insert chart into cell {index}"
            command = commands.SetCellCode(code, model, index, description)

            self.main_window.undo_stack.push(command)

        self.cell2dialog.pop(current)

    def macro_insert_sum(self):
        """Sum up selection area

        The sum is inserted into the cell below the bottom right cell of the
        selection.

        """

        grid = self.main_window.focused_grid
        selection = grid.selection
        shape = grid.model.shape

        (top, left), (bottom, right) = selection.get_grid_bbox(shape)

        if bottom >= shape[0] - 1:
            self.main_window.statusBar().showMessage(
                "ValueError: Target cell is beyond grid limits")
            return

        key = bottom + 1, right, grid.table

        code = f"numpy.sum(eval({repr(selection)}" + \
               f".get_absolute_access_string({shape}, Z)))"

        grid.current = key
        index = grid.currentIndex()
        description = f"Insert sum of {selection} into cell {key}"
        command = commands.SetCellCode(code, grid.model, index, description)

        self.main_window.undo_stack.push(command)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

This file contains interfaces to the xlsx file format.

XlsReader and XlsWriter classed are structured into the following sections:
 * shape
 * code
 * attributes
 * row_heights
 * col_widths
 * macros


**Provides**

 * :func:`wxcolor2rgb`
 * :dict:`wx2qt_fontweights`
 * :dict:`wx2qt_fontstyles`
 * :class:`XlsReader`
 * :class:`XlsWriter`

"""

# from builtins import str, map, object

from collections import defaultdict
import logging
from typing import BinaryIO

try:
    import pycel
    from pycel import excelformula
except ImportError:
    pycel = None

try:
    from openpyxl import load_workbook, worksheet
    from openpyxl.workbook import Workbook
    from openpyxl.cell.cell import (Cell, TYPE_STRING, TYPE_FORMULA,
                                    TYPE_NUMERIC, TYPE_BOOL, TYPE_NULL,
                                    TYPE_INLINE, TYPE_ERROR,
                                    TYPE_FORMULA_CACHE_STRING)
    from openpyxl.utils.cell import column_index_from_string
except ImportError:
    openpyxl = None
    worksheet = None
    Cell = None
    Workbook = None

try:
    from pyspread.lib.attrdict import AttrDict
    from pyspread.lib.selection import Selection
    from pyspread.lib.parsers import spreadsheet_formula_to_code
    from pyspread.model.model import CellAttribute, CodeArray
except ImportError:
    from lib.attrdict import AttrDict
    from lib.selection import Selection
    from lib.parsers import spreadsheet_formula_to_code
    from model.model import CellAttribute, CodeArray


TYPE_DATE = 'd'

BORDERWIDTH_XLSX2PYSU = {"hair": 0, "thin": 1, "medium": 4, "thick": 8}


class XlsxReader:
    """Reads xlsx file from OpenPyXL into a code_array"""

    def __init__(self, xlsx_file: BinaryIO, code_array: CodeArray):
        """
        :param xlsx_file: The xlsx file to be read
        :param code_array: Target code_array

        """

        self.code_array = code_array
        self.wb = load_workbook(xlsx_file)

    def __iter__(self):
        """Iterates over self.xlsx_file, replacing everything in code_array"""

        sheet_attrs = defaultdict(list)

        self._xlsx2shape(self.wb)

        for worksheet_name in self.wb.sheetnames:
            worksheet = self.wb[worksheet_name]
            table = self.wb.sheetnames.index(worksheet_name)

            # Row widths
            self._xlsx2row_heights(worksheet, table)

            # Column widths
            self._xlsx2col_widths(worksheet, table)

            # Merged cells
            self._xlsx2merge_areas(worksheet, table, sheet_attrs)
            self.append_sheet_attributes(table, sheet_attrs)
            sheet_attrs.clear()

            # Cells
            for row_idx, row in enumerate(worksheet.iter_rows()):
                for cell in row:
                    key = row_idx, cell.column-1, table

                    # Code
                    self._xlsx2code(key, cell)

                    # Cell attributes
                    self._xlsx2attributes(key, cell, sheet_attrs)

                    # cell.fill, cell.alignment, cell.border, cell.fill,
                    # cell.font, cell.has_style, cell.hyperlink,
                    # cell.is_date, cell.number_format, cell.protection

                    yield key

            self.append_sheet_attributes(table, sheet_attrs)
            sheet_attrs.clear()

    def append_sheet_attributes(self, table: int, sheet_attrs: dict):
        """Creates selections from attributes in  sheet_attrs and appends them

        :param table: Current table
        :param sheet_attrs: Dict mapping (attr_name, attr values) to key list

        """

        for sheet_attr in sheet_attrs:
            selection = Selection([], [], [], [], sheet_attrs[sheet_attr])
            attr_dict = AttrDict([sheet_attr])
            cell_attr = CellAttribute(selection, table, attr_dict)
            self.code_array.cell_attributes.append(cell_attr)

    @staticmethod
    def xlsx_rgba2rgb(rgb: str) -> (int, int, int):
        """Converts rgba string from xlsx to rgb tuple

        :param rgb: rgba string from xlsx

        """

        if rgb[:2] == "00":
            # Transparent cell
            raise ValueError("Color is transparent")

        r = int(rgb[2:4], 16)
        g = int(rgb[4:6], 16)
        b = int(rgb[6:8], 16)

        return r, g, b

    def _xlsx2shape(self, wb: Workbook):
        """Updates shape in code_array

        :param wb: openpyxl workbook

        """

        sheetnames = self.wb.sheetnames

        max_row = 1
        max_column = 1

        for worksheet_name in sheetnames:
            worksheet = self.wb[worksheet_name]
            if max_row < worksheet.max_row:
                max_row = worksheet.max_row
            if max_column < worksheet.max_column:
                max_column = worksheet.max_column

        self.code_array.macros += f"_sheetnames = {sheetnames}"

        shape = max_row, max_column, len(sheetnames)
        self.code_array.shape = shape

    def _xlsx2row_heights(self, worksheet: worksheet, table: int):
        """Updates row_heights in code_array

        :param worksheet: openpyxl worksheet
        :param table: pyspread table number

        """

        for row in worksheet.row_dimensions:
            height = worksheet.row_dimensions[row].height / 12.8 * 30
            key = row - 1, table
            self.code_array.row_heights[key] = height

    def _xlsx2col_widths(self, worksheet: worksheet, table: int):
        """Updates col_widths in code_array

        :param worksheet: openpyxl worksheet
        :param table: pyspread table number

        """

        for column in worksheet.column_dimensions:
            width = worksheet.column_dimensions[column].width / 11.26 * 100
            key = column_index_from_string(column) - 1, table
            self.code_array.col_widths[key] = width

    @staticmethod
    def _xlsx2merge_areas(worksheet, table, sheet_attrs):
        """Updates merge_areas in sheet_attrs

        :param worksheet: openpyxl worksheet
        :param table: pyspread table number
        :param sheet_attrs: Dict mapping (attr_name, attr values) to key list

        """

        for merged_range in worksheet.merged_cells.ranges:
            area = (merged_range.min_row-1, merged_range.min_col-1,
                    merged_range.max_row-1, merged_range.max_col-1)
            skey = merged_range.min_row-1, merged_range.min_col-1, table
            sheet_attrs[("merge_area", area)].append(skey)

    def _xlsx2code(self, key: (int, int, int),  cell: Cell):
        """Updates code in code_array

        :param key: Pyspread cell key
        :param cell: Cell object from openpyxl

        """

        if pycel is None or cell.data_type in (TYPE_NUMERIC, TYPE_STRING,
                                               TYPE_INLINE, TYPE_DATE,
                                               TYPE_FORMULA_CACHE_STRING):
            code = repr(cell.value)
        elif cell.data_type == TYPE_FORMULA:
            # Convert formula via pycel
            logging.debug(f"Cell {key} is coverted via pycel")
            code = spreadsheet_formula_to_code(cell.value)
        elif cell.data_type == TYPE_BOOL:
            code = "True" if cell.value else "False"
        elif cell.data_type == TYPE_NULL:
            code = None
        elif cell.data_type == TYPE_ERROR:
            code = Exception(str(cell.value))
        else:
            code = repr(cell.value)
            msg = f"Excel data type {cell.data_type} of cell {cell} unknown."
            logging.error(msg)

        logging.debug(f"Cell {key} set to {code}")

        self.code_array.dict_grid[key] = code

    def _xlsx2attributes(self, key: (int, int, int), cell: Cell, sheet_attrs):
        """Updates attributes in code_array

        :param key: Pyspread cell key
        :param cell: Cell object from openpyxl
        :param sheet_attrs: Sheet attribute dict

        """

        skey = key[0], key[1]  # sheet key
        skey_above = key[0] - 1, key[1]
        skey_left = key[0], key[1] - 1

        # Cell background color
        logging.debug("Cell background color")
        try:
            bg_rgb = self.xlsx_rgba2rgb(cell.fill.bgColor.rgb)
            sheet_attrs[("bgcolor", bg_rgb)].append(skey)
        except (ValueError, TypeError):
            pass

        # Text color
        logging.debug("Text color")
        try:
            text_rgb = self.xlsx_rgba2rgb(cell.font.color.rgb)
            sheet_attrs[("textcolor", text_rgb)].append(skey)
        except (AttributeError, ValueError, TypeError):
            pass

        # Text font
        logging.debug("Text font")
        try:
            sheet_attrs[("textfont", cell.font.name)].append(skey)
        except ValueError:
            pass

        # Text size
        logging.debug("Text size")
        try:
            sheet_attrs[("pointsize", cell.font.size)].append(skey)
        except ValueError:
            pass

        # Text weight
        logging.debug("Text weight")
        try:
            if cell.font.bold:
                sheet_attrs[("fontweight", 75)].append(skey)
        except ValueError:
            pass

        # Text style
        logging.debug("Text style")
        try:
            if cell.font.italic:
                sheet_attrs[("fontstyle", 1)].append(skey)
        except ValueError:
            pass

        # Text underline
        logging.debug("Text underline")
        try:
            if cell.font.underline:
                sheet_attrs[("underline", True)].append(skey)
        except ValueError:
            pass

        # Text strikethrough
        logging.debug("Text strikethrough")
        try:
            if cell.font.strike:
                sheet_attrs[("strikethrough", True)].append(skey)
        except ValueError:
            pass

        # Cell alignment
        logging.debug("Cell alignment")

        if cell.alignment.horizontal == "left":
            sheet_attrs[("justification", "justify_left")].append(skey)
        elif cell.alignment.horizontal == "center":
            sheet_attrs[("justification", "justify_center")].append(skey)
        elif cell.alignment.horizontal == "justify":
            sheet_attrs[("justification", "justify_fill")].append(skey)
        elif cell.alignment.horizontal == "right":
            sheet_attrs[("justification", "justify_right")].append(skey)

        if cell.alignment.vertical == "top":
            sheet_attrs[("vertical_align", "align_top")].append(skey)
        elif cell.alignment.vertical == "center":
            sheet_attrs[("vertical_align", "align_center")].append(skey)
        elif cell.alignment.vertical == "bottom":
            sheet_attrs[("vertical_align", "align_bottom")].append(skey)

        # Cell borders
        logging.debug("Cell borders")

        if cell.border.bottom.style is not None:
            width = BORDERWIDTH_XLSX2PYSU[cell.border.bottom.style]
            sheet_attrs[("borderwidth_bottom", width)].append(skey)

        try:
            rgb = cell.border.bottom.color.rgb
            color = self.xlsx_rgba2rgb(rgb)
            sheet_attrs[("bordercolor_bottom", color)].append(skey)
        except (AttributeError, TypeError):
            if cell.border.bottom.style is not None and width is not None:
                color = 0, 0, 0
                sheet_attrs[("bordercolor_bottom", color)].append(skey)

        if cell.border.right.style is not None:
            width = BORDERWIDTH_XLSX2PYSU[cell.border.right.style]
            sheet_attrs[("borderwidth_right", width)].append(skey)

        try:
            rgb = cell.border.right.color.rgb
            color = self.xlsx_rgba2rgb(rgb)
            sheet_attrs[("bordercolor_right", color)].append(skey)
        except (AttributeError, TypeError):
            if cell.border.right.style is not None and width is not None:
                color = 0, 0, 0
                sheet_attrs[("bordercolor_right",
                             color)].append(skey)

        if cell.border.top.style is not None:
            width = BORDERWIDTH_XLSX2PYSU[cell.border.top.style]
            sheet_attrs[("borderwidth_bottom",
                         width)].append(skey_above)

        try:
            rgb = cell.border.top.color.rgb
            color = self.xlsx_rgba2rgb(rgb)
            sheet_attrs[("bordercolor_bottom",
                         color)].append(skey_above)
        except (AttributeError, TypeError):
            if cell.border.top.style is not None and width is not None:
                color = 0, 0, 0
                sheet_attrs[("bordercolor_bottom",
                             color)].append(skey_above)

        if cell.border.left.style is not None:
            width = BORDERWIDTH_XLSX2PYSU[cell.border.left.style]
            sheet_attrs[("borderwidth_right",
                         width)].append(skey_left)

        try:
            rgb = cell.border.left.color.rgb
            color = self.xlsx_rgba2rgb(rgb)
            sheet_attrs[("bordercolor_right",
                         color)].append(skey_left)
        except (AttributeError, TypeError):
            if cell.border.left.style is not None and width is not None:
                color = 0, 0, 0
                sheet_attrs[("bordercolor_right",
                             color)].append(skey_left)



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

Tests for `xlsx.py`

"""

from contextlib import contextmanager
from pathlib import Path
import pytest
import sys

ORIGIN_PATH = Path(__file__).parent.parent.parent


@contextmanager
def insert_path(path):
    sys.path.insert(0, ORIGIN_PATH)
    yield
    sys.path.pop(0)


with insert_path(ORIGIN_PATH):
    from lib.xls_column import xls_column


test_params_excel_column = (
    (0, "A"),
    (1, "B"),
    (2, "C"),
    (23, "X"),
    (24, "Y"),
    (25, "Z"),
    (26, "AA"),
    (650, "YA"),
    (651, "YB"),
    (701, "ZZ"),
    (702, "AAA"),
    (703, "AAB"),
    (16383, "XFD"),
    (2**100+1, "BKOWAWZZAERONYJCMBBOJR"),
)


@pytest.mark.parametrize("column, result", test_params_excel_column)
def test_excel_column(column, result):
    """pytest for `excel_column`"""

    assert xls_column(column) == result



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""
test_grid
=========

Unit tests for grid.py

"""

from contextlib import contextmanager
from os.path import abspath, dirname, join
import sys

import pytest

from PyQt6.QtWidgets import QApplication, QAbstractItemView


PYSPREADPATH = abspath(join(dirname(__file__) + "/.."))
LIBPATH = abspath(PYSPREADPATH + "/lib")


@contextmanager
def insert_path(path):
    sys.path.insert(0, path)
    yield
    sys.path.pop(0)


@contextmanager
def multi_selection_mode(grid):
    grid.clearSelection()
    old_selection_mode = grid.selectionMode()
    grid.setSelectionMode(QAbstractItemView.MultiSelection)
    yield
    grid.setSelectionMode(old_selection_mode)


with insert_path(PYSPREADPATH):
    from ..pyspread import MainWindow
    from ..commands import MakeButtonCell, RemoveButtonCell
    from ..lib.selection import Selection


app = QApplication.instance()
if app is None:
    app = QApplication([])
main_window = MainWindow()
zoom_levels = main_window.settings.zoom_levels


class TestMainWindow:
    """Unit tests for MainWindow in pyspread.py"""

    grid = main_window.grid
    cell_attributes = grid.model.code_array.cell_attributes

    def test_safe_mode(self):
        """Unit test for safe_mode"""

        main_window.safe_mode = True
        main_window.main_window_actions.approve.isEnabled()
        assert main_window.safe_mode

        main_window.safe_mode = False
        assert not main_window.main_window_actions.approve.isEnabled()
        assert not self.grid.model.code_array.result_cache
        assert not main_window.safe_mode

    def test_on_clear_globals(self):
        """Unit test for on_clear_globals"""

        self.grid.model.code_array.result_cache["test"] = "Testres"
        main_window.on_clear_globals()
        assert not self.grid.model.code_array.result_cache



# -*- coding: utf-8 -*-

"""

The functions in this file are extracted from the project qimage2ndarray
https://github.com/hmeine/qimage2ndarray
that has been publisched under the BSD-3-Clause License.


Copyright (c) 2009, Hans Meine <hans_meine@gmx.net>

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""

import numpy
import sys

from PyQt6.QtGui import QImage


if sys.byteorder == 'little':
    _bgra = (0, 1, 2, 3)
else:
    _bgra = (3, 2, 1, 0)


validFormats_8bit = (QImage.Format.Format_Indexed8,
                     QImage.Format.Format_Grayscale8)

validFormats_32bit = (QImage.Format.Format_RGB32,
                      QImage.Format.Format_ARGB32,
                      QImage.Format.Format_ARGB32_Premultiplied)


def PyQt_data(image):
    # PyQt4/PyQt6's QImage.bits() returns a sip.voidptr that supports
    # conversion to string via asstring(size) or getting its base
    # address via int(...):
    return (int(image.bits()), False)


def qimageview(image):
    if not isinstance(image, QImage):
        raise TypeError("image argument must be a QImage instance")

    shape = image.height(), image.width()
    strides0 = image.bytesPerLine()

    format = image.format()
    if format in validFormats_8bit:
        dtype = "|u1"
        strides1 = 1
    elif format in validFormats_32bit:
        dtype = "|u4"
        strides1 = 4
    elif format == QImage.Format.Format_Invalid:
        raise ValueError("qimageview got invalid QImage")
    else:
        msg = (f"qimageview can only handle 8- or 32-bit QImages "
               f"(format was {format})")
        raise ValueError(msg)

    image.__array_interface__ = {
        'shape': shape,
        'typestr': dtype,
        'data': PyQt_data(image),
        'strides': (strides0, strides1),
        'version': 3,
    }

    result = numpy.asarray(image)
    del image.__array_interface__
    return result


def _qimage_or_filename_view(qimage):
    if isinstance(qimage, str):
        qimage = QImage(qimage)
    return qimageview(qimage)


def byte_view(qimage, byteorder='little'):
    """Returns raw 3D view of the given QImage_'s memory.

    This will always be a 3-dimensional numpy.ndarray with dtype numpy.uint8.

    Note that for 32-bit images, the last dimension will be in the
    [B,G,R,A] order (if little endian) due to QImage_'s memory layout
    (the alpha channel will be present for Format_RGB32 images, too).

    For 8-bit (indexed) images, the array will still be 3-dimensional,
    i.e. shape will be (height, width, 1).

    The order of channels in the last axis depends on the `byteorder`,
    which defaults to 'little', i.e. BGRA order.  You may set the
    argument `byteorder` to 'big' to get ARGB, or use None which means
    sys.byteorder here, i.e. return native order for the machine the
    code is running on.

    For your convenience, `qimage` may also be a filename, see
    `Loading and Saving Images`_ in the documentation.

    :param qimage: image whose memory shall be accessed via NumPy
    :type qimage: QImage_
    :param byteorder: specify order of channels in last axis
    :rtype: numpy.ndarray_ with shape (height, width, 1 or 4) and dtype uint8

    """

    raw = _qimage_or_filename_view(qimage)
    result = raw.view(numpy.uint8).reshape(raw.shape + (-1, ))
    if byteorder and byteorder != sys.byteorder:
        result = result[..., ::-1]
    return result


def rgb_view(qimage, byteorder='big'):
    """Returns RGB view of a given 32-bit color QImage_'s memory.

    Similarly to byte_view(), the result is a 3D numpy.uint8 array,
    but reduced to the rgb dimensions (without alpha), and reordered
    (using negative strides in the last dimension) to have the usual
    [R,G,B] order.  The image must have 32 bit pixel size, i.e. be
    RGB32, ARGB32, or ARGB32_Premultiplied.  (Note that in the latter
    case, the values are of course premultiplied with alpha.)

    The order of channels in the last axis depends on the `byteorder`,
    which defaults to 'big', i.e. RGB order.  You may set the argument
    `byteorder` to 'little' to get BGR, or use None which means
    sys.byteorder here, i.e. return native order for the machine the
    code is running on.

    For your convenience, `qimage` may also be a filename, see
    `Loading and Saving Images`_ in the documentation.

    :param qimage: image whose memory shall be accessed via NumPy
    :type qimage: QImage_ with 32-bit pixel type
    :param byteorder: specify order of channels in last axis
    :rtype: numpy.ndarray_ with shape (height, width, 3) and dtype uint8

    """

    if byteorder is None:
        byteorder = sys.byteorder
    bytes = byte_view(qimage, byteorder)
    if bytes.shape[2] != 4:
        msg = ("For rgb_view, the image must have 32 bit pixel size "
               "(use RGB32, ARGB32, or ARGB32_Premultiplied)")
        raise ValueError(msg)

    if byteorder == 'little':
        return bytes[..., :3]  # strip A off BGRA
    else:
        return bytes[..., 1:]  # strip A off ARGB


def alpha_view(qimage):
    """Returns alpha view of a given 32-bit color QImage_'s memory.

    The result is a 2D numpy.uint8 array, equivalent to
    byte_view(qimage)[...,3].  The image must have 32 bit pixel size,
    i.e. be RGB32, ARGB32, or ARGB32_Premultiplied.  Note that it is
    not enforced that the given qimage has a format that actually
    *uses* the alpha channel -- for Format_RGB32, the alpha channel
    usually contains 255 everywhere.

    For your convenience, `qimage` may also be a filename, see
    `Loading and Saving Images`_ in the documentation.

    :param qimage: image whose memory shall be accessed via NumPy
    :type qimage: QImage_ with 32-bit pixel type
    :rtype: numpy.ndarray_ with shape (height, width) and dtype uint8

    """

    bytes = byte_view(qimage, byteorder=None)
    if bytes.shape[2] != 4:
        msg = ("For alpha_view, the image must have 32 bit pixel size "
               "(use RGB32, ARGB32, or ARGB32_Premultiplied)")
        raise ValueError(msg)
    return bytes[..., _bgra[3]]


def _normalize255(array, normalize, clip=(0, 255)):
    if normalize:
        if normalize is True:
            normalize = array.min(), array.max()
            if clip == (0, 255):
                clip = None
        elif numpy.isscalar(normalize):
            normalize = (0, normalize)

        nmin, nmax = normalize

        if nmin:
            array = array - nmin

        if nmax != nmin:
            scale = 255. / (nmax - nmin)
            if scale != 1.0:
                array = array * scale

    if clip:
        low, high = clip
        numpy.clip(array, low, high, array)

    return array


def array2qimage(array, normalize=False):
    """Convert a 2D or 3D numpy array into a 32-bit QImage_.

    The first dimension represents the vertical image axis; the optional
    third dimension is supposed to contain 1-4 channels:

    ========= ===================
    #channels interpretation
    ========= ===================
            1 scalar/gray
            2 scalar/gray + alpha
            3 RGB
            4 RGB + alpha
    ========= ===================

    Scalar data will be converted into corresponding gray RGB triples;
    if you want to convert to an (indexed) 8-bit image instead, use
    `gray2qimage` (which cannot support an alpha channel though).

    The parameter `normalize` can be used to normalize an image's
    value range to 0..255:

    `normalize` = (nmin, nmax):
      scale & clip image values from nmin..nmax to 0..255

    `normalize` = nmax:
      lets nmin default to zero, i.e. scale & clip the range 0..nmax
      to 0..255

    `normalize` = True:
      scale image values to 0..255 (same as passing (array.min(),
      array.max()))

    If `array` contains masked values, the corresponding pixels will
    be transparent in the result.  Thus, the result will be of
    QImage.Format_ARGB32 if the input already contains an alpha
    channel (i.e. has shape (H,W,4)) or if there are masked pixels,
    and QImage.Format_RGB32 otherwise.

    :param array: image data which should be converted (copied) into a QImage_
    :type array: 2D or 3D numpy.ndarray_ or `numpy.ma.array <masked arrays>`_
    :param normalize: normalization parameter (default: no value changing)
    :type normalize: bool, scalar, or pair
    :rtype: QImage_ with RGB32 or ARGB32 format

    """

    if numpy.ndim(array) == 2:
        array = array[..., None]
    elif numpy.ndim(array) != 3:
        tpl = ("array2qimage can only convert 2D or 3D arrays "
               "(got {} dimensions)")
        raise ValueError(tpl.format(numpy.ndim(array)))
    if array.shape[2] not in (1, 2, 3, 4):
        msg = ("array2qimage expects the last dimension to contain exactly "
               "one (scalar/gray), two (gray+alpha), three (R,G,B), or four "
               "(R,G,B,A) channels")
        raise ValueError(msg)

    h, w, channels = array.shape

    hasAlpha = numpy.ma.is_masked(array) or channels in (2, 4)
    fmt = QImage.Format.Format_ARGB32 if hasAlpha else QImage.Format.Format_RGB32

    result = QImage(w, h, fmt)

    array = _normalize255(array, normalize)

    if channels >= 3:
        rgb_view(result)[:] = array[..., :3]
    else:
        rgb_view(result)[:] = array[..., :1]  # scalar data

    alpha = alpha_view(result)

    if channels in (2, 4):
        alpha[:] = array[..., -1]
    else:
        alpha[:] = 255

    if numpy.ma.is_masked(array):
        alpha[:] *= numpy.logical_not(numpy.any(array.mask, axis=-1))

    return result



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

pyspread
========

- Main Python spreadsheet application
- Run this script to start the application.

**Provides**

* MainApplication: Initial command line operations and application launch
* :class:`MainWindow`: Main windows class

"""

import logging
import os
import sys
import traceback

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QApplication

try:
    from pyspread.cli import PyspreadArgumentParser
    from pyspread.main_window import MainWindow

except ImportError:
    from cli import PyspreadArgumentParser
    from main_window import MainWindow


LICENSE = "GNU GENERAL PUBLIC LICENSE Version 3"

os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"


def excepthook(exception_type, exception_value, exception_traceback):
    """Exception hook that prevents pyspread from crashing on exceptions"""

    traceback_msg = "".join(traceback.format_exception(exception_type,
                                                       exception_value,
                                                       exception_traceback))
    print(f"Error: {traceback_msg}\n")


def main():
    """Pyspread main"""

    sys.excepthook = excepthook

    parser = PyspreadArgumentParser()
    args, _ = parser.parse_known_args()

    logging.basicConfig(level=args.loglevel)

    app = QApplication(sys.argv)
    app.setDesktopFileName("io.gitlab.pyspread.pyspread")
    main_window = MainWindow(args.file, default_settings=args.default_settings)

    main_window.show()

    app.exec()

    sys.exit()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

 * :class:`Settings`

"""


from os.path import abspath, dirname, join
from pathlib import Path
from platform import system
from typing import Any

from PyQt6.QtCore import QSettings, QLocale
from PyQt6.QtWidgets import QToolBar, QWidget

try:
    from pyspread.__init__ import VERSION, APP_NAME
except ImportError:
    from __init__ import VERSION, APP_NAME

PYSPREAD_DIRNAME = abspath(join(dirname(__file__), ".."))
PYSPREAD_PATH = Path(PYSPREAD_DIRNAME)
DOC_PATH = PYSPREAD_PATH / "pyspread/share/doc"
TUTORIAL_PATH = DOC_PATH / "tutorial"
MANUAL_PATH = DOC_PATH / "manual"
MPL_TEMPLATE_PATH = PYSPREAD_PATH / 'pyspread/share/templates/matplotlib'
RPY2_TEMPLATE_PATH = PYSPREAD_PATH / 'pyspread/share/templates/rpy2'
PLOT9_TEMPLATE_PATH = PYSPREAD_PATH / 'pyspread/share/templates/plotnine'
ICON_PATH = PYSPREAD_PATH / 'pyspread/share/icons'
ACTION_PATH = ICON_PATH / 'actions'
STATUS_PATH = ICON_PATH / 'status'
CHARTS_PATH = ICON_PATH / 'charts'
WEB_URL = "https://pyspread.gitlab.io"  # Official Web page

AUTHORS = """Martin Manns
             Jason Sexauer
             Vova Kolobok
             mgunyho
              Pete Morgan
             Ari Caldeira
             kirinokirino
             Seongyong Park
             Merikei"""


class Settings:
    """Contains all global application states."""

    # Note that `safe_mode` is not listed here but inside
    # :class:`model.model.DataArray`

    icon_size = 24, 24

    """Size of toolbar icons"""

    widget_names = ["main_window", "main_toolbar", "find_toolbar",
                    "format_toolbar", "macro_toolbar", "entry_line",
                    "entry_line_dock"]

    """Names of widgets with persistant states"""

    shape = 1000, 100, 3
    """Default shape of initial grid (rows, columns, tables)"""

    maxshape = 1000000, 100000, 100
    """Maximum shape of the grid"""

    changed_since_save = False
    """If `True` then File actions trigger a dialog"""

    last_file_input_path = Path.home()
    """Initial :class:`~pathlib.Path` for opening files"""

    last_file_output_path = Path.home()
    """Initial :class:`~pathlib.Path` for saving files"""

    last_file_import_path = Path.home()
    """Initial :class:`~pathlib.Path` for importing files"""

    last_file_export_path = Path.home()
    """Initial :class:`~pathlib.Path` for exporting files"""

    max_file_history = 5
    """Maximum number of files in file history"""

    file_history = []
    """Files in file history"""

    entry_line_font_family = "Monospace"
    """Font family for entry line widget"""

    macro_editor_font_family = "Monospace"
    """Font family for macro editor widget"""

    digest_types = None
    """List of default digest types for preprocessing values from CSV import"""

    highlighter_limit = 1000000
    """Maximum length of code, for which the netry line enables highlighting"""

    border_choice = "All borders"
    """The state of the border choice button"""

    timeout = 1000
    """Timeout for cell calculations in milliseconds"""

    refresh_timeout = 1000
    """Timeout for frozen cell updates in milliseconds"""

    signature_key = None
    """Key for signing save files"""

    font_sizes = (6, 8, 10, 12, 14, 16, 18, 20, 24, 28, 32)
    """Sizes"""

    default_row_height = 30
    default_column_width = 100

    zoom_levels = (0.4, 0.5, 0.6, 0.7, 0.8, 1.0,
                   1.2, 1.4, 1.6, 1.8, 2.0, 2.5, 3.0, 3.5, 4.0, 5.0, 6.0, 8.0)

    print_zoom = None

    show_frozen = False
    """If `True` then frozen cell background is striped"""

    find_dialog_state = None
    """Find dialog state - needs to be stored when dialog is closed"""

    default_encoding = "utf-8"
    """Default encoding for exporting files (e.g. CSV)"""

    currency_iso_code = QLocale.system().currencySymbol(
        QLocale.CurrencySymbolFormat.CurrencyIsoCode)
    """Default currency for Money shortcut"""

    encodings = (
        "ascii", "big5", "big5hkscs", "cp037", "cp424", "cp437",
        "cp500", "cp720", "cp737", "cp775", "cp850", "cp852", "cp855", "cp856",
        "cp857", "cp858", "cp860", "cp861", "cp862", "cp863", "cp864", "cp865",
        "cp866", "cp869", "cp874", "cp875", "cp932", "cp949", "cp950",
        "cp1006", "cp1026", "cp1140", "cp1250", "cp1251", "cp1252", "cp1253",
        "cp1254", "cp1255", "cp1256", "cp1257", "cp1258", "euc-jp",
        "euc-jis-2004", "euc-jisx0213", "euc-kr", "gb2312", "gbk", "gb18030",
        "hz", "iso2022-jp", "iso2022-jp-1", "iso2022-jp-2", "iso2022-jp-2004",
        "iso2022-jp-3", "iso2022-jp-ext", "iso2022-kr", "latin-1", "iso8859-2",
        "iso8859-3", "iso8859-4", "iso8859-5", "iso8859-6", "iso8859-7",
        "iso8859-8", "iso8859-9", "iso8859-10", "iso8859-13", "iso8859-14",
        "iso8859-15", "iso8859-16", "johab", "koi8-r", "koi8-u",
        "mac-cyrillic", "mac-greek", "mac-iceland", "mac-latin2", "mac-roman",
        "mac-turkish", "ptcp154", "shift-jis", "shift-jis-2004",
        "shift-jisx0213", "utf-32", "utf-32-be", "utf-32-le", "utf-16",
        "utf-16-be", "utf-16-le", "utf-7", "utf-8", "utf-8-sig",
    )
    """Encodings for importing and exporting files (e.g. CSV or SVG)"""

    sniff_size = 65536
    """Number of bytes for csv sniffer
       sniff_size should be larger than 1st+2nd line
    """

    # Status bar cell result summation
    show_statusbar_sum = True

    def __init__(self, parent: QWidget, reset_settings: bool = False):
        """
        :param parent: Parent widget, normally main window
        :param reset_settings: Do not restore saved settings

        """

        super().__setattr__("parent", parent)
        super().__setattr__("reset_settings", reset_settings)

    def __setattr__(self, key: str, value: Any):
        """
        Overloads __setattr__ to ensure that only existing attributes are set

        :param key: Setting attribute key
        :param value: New setting value

        """

        if not hasattr(self, key):
            raise AttributeError(f"{self} has no attribute {key}.")
        super().__setattr__(key, value)

    def add_to_file_history(self, filename: Path):
        """Adds new file to history

        :param value: File name to be added to history

        """

        self.file_history = [f for f in self.file_history if f != filename]
        self.file_history.insert(0, filename)
        self.file_history = self.file_history[:self.max_file_history]

    def reset(self):
        """Reset to defaults"""

        cls_attrs = (attr for attr in dir(self)
                     if (not attr.startswith("__")
                         and attr not in ("reset", "parent", "save",
                                          "restore", "default_settings")))
        for cls_attr in cls_attrs:
            setattr(self, cls_attr, getattr(Settings, cls_attr))

    def save(self):
        """Saves application state to QSettings"""

        if system() == "Darwin":
            settings = QSettings(APP_NAME+".gitlab.io", APP_NAME)
        else:
            settings = QSettings(APP_NAME, APP_NAME)

        # Application state

        # Do not store the actual filename. Otherwise, after saving and closing
        # File -> Save would overwrite the last saved file.

        if self.last_file_input_path is not None:
            settings.setValue("last_file_input_path",
                              self.last_file_input_path)
        if self.last_file_import_path is not None:
            settings.setValue("last_file_import_path",
                              self.last_file_import_path)
        if self.last_file_export_path is not None:
            settings.setValue("last_file_export_path",
                              self.last_file_export_path)
        settings.setValue("max_file_history", self.max_file_history)
        settings.value("file_history", [], 'QStringList')
        if self.file_history:
            settings.setValue("file_history", self.file_history)
        settings.setValue("timeout", self.timeout)
        settings.setValue("refresh_timeout", self.refresh_timeout)
        settings.setValue("signature_key", self.signature_key)
        settings.setValue("show_statusbar_sum", self.show_statusbar_sum)

        # GUI state
        for widget_name in self.widget_names:

            if widget_name == "main_window":
                widget = self.parent
            else:
                widget = getattr(self.parent, widget_name)

            # geometry
            geometry_name = widget_name + '/geometry'
            try:
                settings.setValue(geometry_name, widget.saveGeometry())
            except AttributeError:
                pass

            # state
            widget_state_name = widget_name + '/windowState'
            try:
                settings.setValue(widget_state_name, widget.saveState())
            except AttributeError:
                pass

            if isinstance(widget, QToolBar):
                toolbar_visibility_name = widget_name + '/visibility'
                settings.value(toolbar_visibility_name, [], bool)
                settings.setValue(toolbar_visibility_name,
                                  [a.isVisible() for a in widget.actions()])

            if widget_name == "entry_line_dock":
                settings.setValue("entry_line_isvisible", widget.isVisible())

        settings.sync()

    def restore(self):
        """Restores application state from QSettings"""

        def qt_bool(value):
            """Converts Qt setting string for bool into Python bool"""
            if value == "true":
                return True
            else:
                return False

        if self.reset_settings:
            return

        if system() == "Darwin":
            settings = QSettings(APP_NAME+".gitlab.io", APP_NAME)
        else:
            settings = QSettings(APP_NAME, APP_NAME)

        def setting2attr(setting_name, attr=None, mapper=None):
            """Sets attr to mapper(<Setting from setting_name>)"""

            value = settings.value(setting_name)
            if value is None:
                return
            if attr is None:
                attr = setting_name
            if mapper is None:
                def mapper(x):
                    return x
            setattr(self, attr, mapper(value))

        # Application state

        setting2attr("last_file_input_path")
        setting2attr("last_file_import_path")
        setting2attr("last_file_export_path")
        setting2attr("max_file_history", mapper=int)
        setting2attr("file_history")
        setting2attr("timeout", mapper=int)
        setting2attr("refresh_timeout", mapper=int)
        setting2attr("signature_key")
        setting2attr("show_statusbar_sum", mapper=qt_bool)

        # GUI state

        for widget_name in self.widget_names:
            geometry_name = widget_name + '/geometry'
            widget_state_name = widget_name + '/windowState'

            if widget_name == "main_window":
                widget = self.parent
            else:
                widget = getattr(self.parent, widget_name)

            geometry = settings.value(geometry_name)
            if geometry:
                widget.restoreGeometry(geometry)
            widget_state = settings.value(widget_state_name)
            if widget_state:
                widget.restoreState(widget_state)

            if isinstance(widget, QToolBar):
                toolbar_visibility_name = widget_name + '/visibility'
                visibility = settings.value(toolbar_visibility_name)
                if visibility is not None:
                    for is_visible, action in zip(visibility,
                                                  widget.actions()):
                        action.setVisible(is_visible in ['true', True])
                manager_button = widget.widgetForAction(widget.actions()[-1])
                manager_button.menu().update_checked_states()

            if widget_name == "entry_line_dock" \
               and settings.value("entry_line_isvisible") is not None:
                visible = settings.value("entry_line_isvisible") in ['true',
                                                                     True]
                widget.setVisible(visible)
